# -*- coding: utf8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import tempfile
from unittest import TestCase

import filternaut
from django import VERSION as DJANGO_VERSION
from django.utils.datastructures import MultiValueDict
from filternaut import Filter, Optional
from filternaut.filters import (CharFilter, ChoiceFilter, FilePathFilter,
                                RegexFilter, ComboFilter, FieldFilter)
from tests.util import NopeFilter, flatten_qobj
from django.forms import CharField, IntegerField


class FilterTests(TestCase):

    def test_kwargless_simple_instantiation(self):
        f = Filter('fieldname')
        assert f.source == 'fieldname'
        assert f.dest == 'fieldname'

    def test_kwarg_instantiation(self):
        f = Filter(dest='fieldname')
        assert f.source == 'fieldname'
        assert f.dest == 'fieldname'

        f = Filter(source='sourcename', dest='fieldname')
        assert f.source == 'sourcename'
        assert f.dest == 'fieldname'

    def test_getsourcevalue_does_what_it_says_on_the_tin(self):
        f = Filter('fieldname')
        bar = f.get_source_value('foo', {'foo': 'bar'})
        assert bar == 'bar'

    def test_getsourcevalue_raises_keyerrors(self):
        f = Filter('fieldname')
        with self.assertRaises(KeyError):
            f.get_source_value('legume', {'foo': 'bar'})

    def test_required_filter_is_satisfied_by_one_key_being_present(self):
        f = Filter('fieldname', required=True, lookups=['gte'])
        data = {
            'fieldname__gte': 'foo',
            'fieldname__lte': 'bar'}
        f.parse(data)
        assert not f.errors

    def test_required_filter_requires_at_least_one_key(self):
        """
        One filter can have many possible sources due to the different lookup
        affixes. One of them being present should be enough to satisfy a
        'required' field.
        """
        f = Filter('fieldname', required=True, lookups=['contains'])
        data = {
            'fieldname__gte': 'foo',
            'fieldname__lte': 'bar'}
        f.parse(data)
        assert f.errors
        assert 'fieldname' in f.errors

    def test_lookups_can_be_provided_as_a_string(self):
        f = Filter('fieldname', lookups='contains')
        assert f.lookups == ['contains']

    def test_lookups_can_be_provided_as_a_comma_separated_string(self):
        f = Filter('fieldname', lookups='lt,lte,gt,gte')
        assert f.lookups == ['lt', 'lte', 'gt', 'gte']

    def test_negating_filters(self):
        f = Filter('fieldname')
        assert f.negate is False
        f = ~f
        assert f.negate is True
        f = ~f
        assert f.negate is False

    def test_one_source_to_many_dests(self):
        filters = (
            Filter(source='search', dest='username') |
            Filter(source='search', dest='email') |
            Filter(source='search', dest='name'))
        filters.parse(dict(search='mynbaev-scheiner'))
        expected = dict(
            username='mynbaev-scheiner',
            email='mynbaev-scheiner',
            name='mynbaev-scheiner')
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual

    def test_many_sources_one_dest(self):
        filters = (
            Filter(source='nominee_1', dest='field') |
            Filter(source='nominee_2', dest='field') |
            Filter(source='nominee_3', dest='field'))
        filters.parse(dict(
            nominee_1='one',
            nominee_2='two',
            nominee_3='three'))
        expected = dict(field='three')  # last one wins
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual

    def test_multivaluedict_as_source(self):
        """
        We don't expect multiple values to be pulled from a multi-value source
        without explicit instruction
        """
        filters = Filter('name')
        data = MultiValueDict(dict(name=['foo', 'bar']))
        filters.parse(data)
        expected = dict(name='bar')  # last one wins
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual

    def test_multivaluedict_as_source_when_many_values_required(self):
        """
        Having a dest ending in '__in' /does/ constitute explicit instruction
        to pull multiple values from a source.
        """
        filters = Filter('field', lookups=['in'])
        data = MultiValueDict(dict(field=['foo', 'bar']))
        filters.parse(data)
        expected = dict(field__in=['foo', 'bar'])
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual

    def test_multivaluedict_not_used_for_nonlisty_filtering(self):
        filters = Filter('field', lookups=['exact', 'in'])

        data = MultiValueDict(dict(field=['foo', 'bar']))
        filters.parse(data)
        expected = dict(field='bar')  # last one wins
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual

        data = MultiValueDict(dict(field__in=['foo', 'bar']))
        filters.parse(data)
        expected = dict(field__in=['foo', 'bar'])
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual


class TestSourceDestPairCreation(TestCase):

    def test_no_lookup_same_source_and_dest(self):
        filter = Filter('field', lookups=[None])
        expected = [('field', 'field'), ]
        actual = list(filter.source_dest_pairs())
        assert expected == actual

    def test_gte_lookup_same_source_and_dest(self):
        filter = Filter('field', lookups=['gte'])
        expected = [
            ('field__gte', 'field__gte'),
            ('field', 'field__gte')]
        actual = list(filter.source_dest_pairs())
        assert sorted(expected) == sorted(actual)

    def test_no_lookup_differing_source_and_dest(self):
        filter = Filter(source='something', dest='field', lookups=[None])
        expected = [('something', 'field'), ]
        actual = list(filter.source_dest_pairs())
        assert expected == actual

    def test_lte_lookup_differing_source_and_dest(self):
        filter = Filter(source='something', dest='field', lookups=['lte'])
        expected = [
            ('something__lte', 'field__lte'),
            ('something', 'field__lte')]
        actual = list(filter.source_dest_pairs())
        assert sorted(expected) == sorted(actual)


class ParsingTests(TestCase):

    def test_parse_simple_source(self):
        filter = Filter('word')
        data = {
            'word': 'a_value',
            'unrelated_key': ''}
        filter.parse(data)

        assert not filter.errors
        assert 'word' in filter._filters
        assert filter._filters['word'] == 'a_value'

    def test_parse_source_with_lookups(self):
        filter = Filter('word', lookups=['gte', 'gt', 'lte', 'lt'])
        data = {
            'word__gte': 1,
            'word__gt': 2,
            'word__lte': 3,
            'word__lt': 4}

        filter.parse(data)

        assert not filter.errors
        for key in data:
            assert key in filter._filters
        for key, value in data.items():
            assert filter._filters[key] == value
        assert filter.valid

    def test_more_extensive_parsing(self):
        filters = (
            Filter('field_1', lookups=['contains', 'icontains']) |
            Filter(source='field_2', dest='field_2a') |
            Filter(source='field_3', dest='f3', lookups=['lt', 'gt']) |
            Filter(source='id', dest='relationship__spanning__id'))
        data = {
            'field_1__contains': 'first',
            'field_2': 'second',
            'field_3__lt': 'third',
            'field_3__gt': 'fourth',
            'id': 'fifth'}
        expected = {
            'field_1__contains': 'first',
            'field_2a': 'second',
            'f3__lt': 'third',
            'f3__gt': 'fourth',
            'relationship__spanning__id': 'fifth'}
        filters.parse(data)
        actual = dict(flatten_qobj(filters.Q))
        assert expected == actual
        assert filters.valid


class FieldFilterTests(TestCase):

    def test_choicefilter_choices(self):
        choices = (
            ('alai', 'Alai'),
            ('petra', 'Petra'))

        filter = ChoiceFilter(choices=choices, dest='name')
        filter.parse({'name': 'bean'})
        assert filter.errors

        filter = ChoiceFilter(choices=choices, dest='name')
        filter.parse({'name': 'alai'})
        assert not filter.errors

    def test_choicefilter_instantiation_styles(self):
        fieldname = 'fieldname'
        choices = ['choice1', 'choice2']

        filters = [
            # no kwargs
            ChoiceFilter(fieldname, choices),
            # arg fieldname, kwarg choices
            ChoiceFilter(fieldname, choices=choices),
            # all kwargs
            ChoiceFilter(source=fieldname, dest=fieldname, choices=choices)]

        for filter in filters:
            assert filter.source == fieldname
            assert filter.field.choices == choices

    def test_regexfilter(self):
        filter = RegexFilter('fieldname', r'\d+')
        filter.parse({'fieldname': 'alpha'})
        assert filter.errors
        assert 'fieldname' in filter.errors

        filter = RegexFilter('fieldname', r'\d+')
        filter.parse({'fieldname': '1008346'})
        assert not filter.errors
        assert filter.valid

    def test_filepathfilter(self):
        # this isn't a great test, since we just instantiate the filter.
        # however, proper testing requires monkeying about changing the test
        # system's filesystem; and, being a subclass of choicefilter, that
        # class' tests should cover us.
        path = tempfile.gettempdir()
        FilePathFilter('fieldname', path)

    def test_combofilter(self):
        filter = ComboFilter('fieldname', fields=[
            CharField(min_length=2, max_length=4),
            CharField(min_length=3, max_length=6)
        ])
        invalids = '', 'a', 'aa', 'aaaaa', 'aaaaaa', 'aaaaaaa'
        valids = 'aaa', 'aaaa'

        for invalid in invalids:
            filter.parse({'fieldname': invalid})
            assert filter.errors
            assert 'fieldname' in filter.errors

        for valid in valids:
            filter.parse({'fieldname': valid})
            assert not filter.errors
            assert filter.valid

    def test_multivaluefilter(self):
        # TODO this field is relatively complex. have not written a test for it
        # yet.
        pass

    def test_instantiate_all_filterfields_without_special_args(self):
        # ChoiceFilter, FilePathFilter, and RegexFilter have their own specific
        # tests, because their fields' constructors, and therefore their own
        # constructors, require additional arguments
        filterfields = (
            'BooleanFilter', 'CharFilter', 'DateFilter',
            'DateTimeFilter', 'DecimalFilter', 'EmailFilter', 'FloatFilter',
            'ImageFilter', 'IntegerFilter', 'MultipleChoiceFilter',
            'NullBooleanFilter', 'SlugFilter', 'SplitDateTimeFilter',
            'TimeFilter', 'TypedChoiceFilter', 'URLFilter',
            'GenericIPAddressFilter', 'TypedMultipleChoiceFilter'
        )

        for f in filterfields:
            klass = getattr(filternaut.filters, f)
            klass('fieldname')

    def test_listlike_values_cleaned_individually(self):
        f = CharFilter('fieldname', lookups='exact,in')
        data = MultiValueDict({
            'fieldname': ['single value'],  # MVDict treats 1-list as single
            'fieldname__in': ['multiple', 'values'],
        })
        expected = {
            'fieldname': 'single value',
            'fieldname__in': ['multiple', 'values']
        }
        f.parse(data)
        actual = dict(flatten_qobj(f.Q))
        assert expected == actual


class DefaultValueTests(TestCase):

    def test_default_value_used_if_no_sourcedata_found(self):
        filters = Filter('count', lookups=['lte', 'gte'], default=3)
        filters.parse({})  # no value for 'count'

        expected = {'count__exact': 3}
        actual = dict(flatten_qobj(filters.Q))

        assert expected == actual

    def test_default_lookup_is_exact(self):
        f = Filter('count', default=3)
        assert f.default_lookup == 'exact'

    def test_other_lookups_ignored_when_default_used(self):
        filters = Filter('count', lookups=['lte', 'gte'], default=3)
        filters.parse({})  # no value for 'count'

        keys = dict(flatten_qobj(filters.Q)).keys()

        assert 'lte' not in keys
        assert 'gte' not in keys

    def test_default_lookup_type_can_be_changed(self):
        filters = Filter('count', lookups=['lte', 'gte'], default=3,
                         default_lookup='gt')
        filters.parse({})  # no value for 'count'

        expected = {'count__gt': 3}
        actual = dict(flatten_qobj(filters.Q))

        assert expected == actual

    def test_default_ignored_if_sourcedata_found(self):
        filters = Filter('count', lookups=['lte', 'gte'], default=3,
                         default_lookup='foobarbaz')
        filters.parse({'count__gte': 4})

        expected = {'count__gte': 4}
        actual = dict(flatten_qobj(filters.Q))

        assert expected == actual

    def test_default_can_be_callable(self):
        filters = Filter('count', lookups=['lte', 'gte'], default=lambda: 3)
        filters.parse({})  # no value for 'count'

        expected = {'count__exact': 3}
        actual = dict(flatten_qobj(filters.Q))

        assert expected == actual


class OptionalTests(TestCase):

    def setUp(self):
        self.filters = Optional(
            Filter('one', required=True),
            Filter('two'),
            Filter('three', required=True))
        self.unrelated = Filter('ten') & (Filter('eleven') | Filter('twelve'))

    def test_no_values_present(self):
        data = dict()  # no data
        self.filters.parse(data)
        assert self.filters.valid

    def test_all_values_present(self):
        data = dict(one=1, two=2, three=3)
        self.filters.parse(data)
        assert self.filters.valid

    def test_only_required_values_present(self):
        data = dict(one=1, three=3)
        self.filters.parse(data)
        assert self.filters.valid

    def test_only_nonrequired_values_present(self):
        data = dict(two=2)
        self.filters.parse(data)
        assert not self.filters.valid
        assert 'one' in self.filters.errors
        assert 'three' in self.filters.errors
        assert '__all__' in self.filters.errors

    def test_one_required_filter_missing(self):
        data = dict(two=2, three=3)
        self.filters.parse(data)
        assert not self.filters.valid
        assert 'one' in self.filters.errors
        assert 'three' not in self.filters.errors
        assert '__all__' in self.filters.errors

    def test_ANDed_with_unrelated(self):
        filters = self.filters & self.unrelated

        data = dict()  # no data
        filters.parse(data)
        assert filters.valid

        data = dict(one=1, three=3)  # all required data
        filters.parse(data)
        assert filters.valid

        data = dict(ten=10)  # only some unrelated
        filters.parse(data)
        assert filters.valid

        data = dict(ten=10, eleven=11, twelve=12)  # only all unrelated
        filters.parse(data)
        assert filters.valid

    def test_validation_errors_are_not_silenced(self):
        filters = Optional(
            NopeFilter('a', required=True),
            NopeFilter('b', required=True),
            NopeFilter('c'))
        data = dict(a=1, b=2, c=3)
        filters.parse(data)
        assert 'a' in filters.errors
        assert 'b' in filters.errors
        assert 'c' in filters.errors

        data = dict()
        filters.parse(data)
        assert not filters.errors


def boolean_tests():
    filter = filternaut.filters.BooleanFilter('approved')

    valid_values = '1', '0', 'true', 'false', 'True', 'False', True, False
    for value in valid_values:
        filter.parse(dict(approved=value))
        assert filter.valid


class MultiValueWithNoneTests(TestCase):

    def get_output_of_field(self, isnull=False, **data):
        filter = FieldFilter('rank', lookups='in', none_to_isnull=isnull,
                             field=IntegerField(required=False))
        data = MultiValueDict(data)
        filter.parse(data)
        assert filter.valid
        return dict(filter.Q.children)

    def test_enabled_with_regular_value(self):
        actual = self.get_output_of_field(rank=['1', '2', '3'], isnull=True)
        expected = {'rank__in': [1, 2, 3]}
        assert actual == expected

    def test_enabled_with_none_value_and_regulars(self):
        actual = self.get_output_of_field(rank=['1', '2', '3', ''], isnull=True)
        expected = {'rank__in': [1, 2, 3], 'rank__isnull': True}
        assert actual == expected

    def test_enabled_with_only_none_value(self):
        actual = self.get_output_of_field(rank=[''], isnull=True)
        expected = {'rank__isnull': True}
        assert actual == expected

    def test_disabled_with_regular_value(self):
        actual = self.get_output_of_field(rank=['1', '2', '3'], isnull=False)
        expected = {'rank__in': [1, 2, 3]}
        assert actual == expected

    def test_disabled_with_none_value_and_regulars(self):
        actual = self.get_output_of_field(rank=['1', '2', '3', ''], isnull=False)
        expected = {'rank__in': [1, 2, 3, None]}
        assert actual == expected

    def test_disabled_with_only_none_value(self):
        actual = self.get_output_of_field(rank=[''], isnull=False)
        expected = {'rank__in': [None]}
        assert actual == expected



# -*- coding: utf8 -*-

from __future__ import unicode_literals
from copy import deepcopy
from operator import and_, or_

from django.core.exceptions import ValidationError
from django.db.models import Q
import six

from filternaut.tree import Tree, Leaf


class FilterTree(Tree):
    """
    FilterTrees instances are the result of ORing or ANDing Filter instances,
    or other FilterTree instances, together.

    FilterTree provides a simple API for dealing with a set of filters. They
    may also be iterated over to gain access to the Filter instances directly.
    """

    def __init__(self, negate=False, *args, **kwargs):
        self.negate = negate
        super(FilterTree, self).__init__(*args, **kwargs)

    def __invert__(self):
        inverted = deepcopy(self)
        inverted.negate = not self.negate
        return inverted

    @property
    def valid(self):
        """
        A boolean indicating whether all filters parsed successfully. Cannot be
        read before ``parse()`` has been called.
        """
        return not self.errors

    @property
    def errors(self):
        """
        A dictionary of errors met while parsing. The errors are keyed by their
        field's source value. Each key's value is a list of errors.

        ``errors`` cannot be read before ``parse()`` has been called.
        """
        errors = self.left.errors
        errors.update(self.right.errors)
        return errors

    @property
    def Q(self):
        """
        A Django "Q" object, built by combining input data with the filter
        definitions. Cannot be read before ``parse()`` has been called.
        """
        q = self.operator(self.left.Q, self.right.Q)
        return ~q if self.negate else q

    def parse(self, data):
        """
        Ask all filters to look through ``data`` and thereby configure
        themselves.
        """
        for fltr in self:
            fltr.parse(data)


class Optional(FilterTree):
    """
    Filters included underneath Optional have their required=True configuration
    ignored as long as all those filters are missing. If some but not all are
    present, then required=True is observed, and those filters that are missing
    become invalid. If all filters under Optional have valid source data, they
    are valid as a whole.

    This is useful for situations where you want to require one field if
    another is present. For example, requiring ``last_name`` if ``first_name``
    is present, but also allowing neither. In this case, you would mark both
    with required=True, and wrap them in Optional:

        Optional(
            Filter('first_name', required=True),
            Filter('last_name', required=True))

    Generally, filters underneath Optional will have required=True, however it
    isn't necessary; consider adding 'middle name' to the above example.
    """

    tree_class = FilterTree

    def __init__(self, left, *rest):
        operator = and_  # OR does not make sense with group-require
        if not rest:
            # TODO if left is a tree, walk it instead of complaining
            raise ValueError("Optional has no effect on a single filter")
        right = six.moves.reduce(and_, rest)
        super(Optional, self).__init__(False, operator, left, right)

    @property
    def errors(self):
        errors = super(Optional, self).errors
        filters = list(self)
        missing = [f.missing for f in filters if f.required]
        present = [f.dict for f in filters]

        # some filters have values, but not all required filters have values.
        if any(missing) and any(present) and not all(present):
            # insert an additional error
            sources = sorted([f.source for f in filters])
            if '__all__' not in errors:
                errors['__all__'] = []
            errors['__all__'].append(
                'If any of {} are provided, all must be '
                'provided'.format(', '.join(sources)))
        else:
            for f in filters:
                if f.required and f.missing:
                    del errors[f.source]

        return errors


class Filter(Leaf):
    """
    A Filter instance builds a django.db.models.Q object by pulling a value
    from arbitrary native data, e.g. a set of query params.

    It can be ORed or ANDed together with other Filter instances in the same
    way Q objects can.
    """

    #: Filters combine into FilterTree instances
    tree_class = FilterTree

    def __init__(self, dest, **kwargs):
        self.dest = dest
        self.source = kwargs.get('source', dest)
        self.lookups = kwargs.get('lookups', ['exact'])
        self.multivalue_lookups = kwargs.get('multivalue_lookups', ['in'])
        self.required = kwargs.get('required', False)
        self.negate = kwargs.get('negate', False)
        self.none_to_isnull = kwargs.get('none_to_isnull', False)

        # None is a valid default -- consider exclude(groups=None) -- so use
        # the absence or presence of self.default to indicate whether a default
        # should be used.
        if 'default' in kwargs:
            self.default = kwargs['default']
            self.default_lookup = kwargs.get('default_lookup', 'exact')

        self._filters = {}
        self._errors = {}
        self.parsed = False
        self.missing = False

        # accept lookups as a comma-separated string.
        if isinstance(self.lookups, six.string_types):
            self.lookups = self.lookups.split(',')

    def __invert__(self):
        """
        Invert the sense of this filter.
        """
        inverted = deepcopy(self)
        inverted.negate = not self.negate
        return inverted

    def parse(self, data):
        """
        Look through the provided dict-like data for keys which match this
        Filter's source.  This includes keys containg lookup affixes such as
        'contains' or 'lte'.

        Once this method has been called, the ``errors``, ``valid`` and ``Q``
        attributes become usable.
        """
        source_pairs = self.source_value_pairs(data)
        dest_pairs, errors = self.dest_value_pairs(source_pairs)

        # handle default value
        if not source_pairs and hasattr(self, 'default'):
            dest_pairs = (self.default_dest_value_pair(), )

        # if required, check if satisfied
        if not source_pairs and self.required:
            self.missing = True
            if self.source not in errors:
                errors[self.source] = []
            errors[self.source].append('This field is required')
        else:
            # this allows a later parse() to undo an earlier missing=True
            self.missing = False

        self.parsed = True
        self._filters = dict(dest_pairs)
        self._errors = errors

    def clean(self, value):
        """
        Validate and normalise ``value`` for use in filtering. This
        implementation is a no-op; subclasses may do more work here.
        """
        return value

    def source_dest_pairs(self):
        """
        For each lookup in self.lookups, such as 'contains' or 'lte', combine
        it with this field's source and dest, returning e.g.
        (username__contains, account_name__contains)

        If any lookup is None, that pair becomes (source, dest)

        If there is only one lookup, two pairs are listed containing the source
        both with and without the lookup. This allows source data to omit the
        lookup from the key, e.g. providing 'email' to the filter
        Filter('email', lookups=['iexact']).
        """
        pairs = []
        for lookup in self.lookups:
            if lookup in (None, 'exact'):
                source = self.source
                dest = self.dest
            else:
                source = '{}__{}'.format(self.source, lookup)
                dest = '{}__{}'.format(self.dest, lookup)
            pairs.append((source, dest))

        # allow source data to omit the lookup if only one lookup listed.
        if len(pairs) == 1:
            lookup = self.lookups[0]
            if lookup not in (None, 'exact'):
                dest = '{}__{}'.format(self.dest, self.lookups[0])
                pairs.append((self.source, dest))

        return pairs

    def source_value_pairs(self, data):
        """
        Return a list of two-tuples containing valid sources for this filter --
        made by combining this filter's source and the various lookups -- and
        their values, as pulled from the data handed to parse.

        Sources with no found value are excluded.
        """
        pairs = []
        for source, dest in self.source_dest_pairs():
            try:
                many = any(dest.endswith("__%s" % x)
                           for x in self.multivalue_lookups)
                value = self.get_source_value(source, data, many)
                pairs.append((source, value))
            except KeyError:
                pass
        return pairs

    def dest_value_pairs(self, sourcevalue_pairs):
        """
        Return two values:
            - A list of two-tuples containing dests (ORM relation/field names)
              and their values.
            - A dictionary of errors, keyed by the source which they originated
              from.
        """
        sourcedest_map = dict(self.source_dest_pairs())
        pairs = []
        errors = {}
        for source, value in sourcevalue_pairs:
            try:
                value = self.clean(value)
                dest = sourcedest_map[source]
                pairs.append((dest, value))
            except ValidationError as ex:
                errors[source] = ex.messages
        return pairs, errors

    def default_dest_value_pair(self):
        """
        Construct a default dest/value pair to be used if no source data was
        found during parsing (and if this filter has default=True).
        """
        dest = '{}__{}'.format(self.dest, self.default_lookup)
        default = self.default() if callable(self.default) else self.default
        return (dest, default)

    def get_source_value(self, key, data, many=False):
        """
        Pull ``key`` from ``data``.

        When ``many`` is True, a list of values is returned. Otherwise, a
        single value is returned.
        """
        if many is False:
            return data[key]
        elif hasattr(data, 'getlist'):  # Django querydict, multivaluedict
            if key not in data:
                raise KeyError(repr(key))
            return data.getlist(key)
        else:
            # only a single value, but many=True, so return as list.
            return [data[key]]

    @property
    def dict(self):
        """
        A dictionary representation of this Filter's filter configuration.
        Cannot be read before ``parse()`` has been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on this filter before "
                "accessing this attribute")
        return self._filters

    @property
    def valid(self):
        """
        A boolean indicating whether this Filter registered any errors during
        parsing. Raises a ValueError if ``parse()`` has not been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on this filter before checking validity")
        return not self._errors

    @property
    def errors(self):
        """
        A dictionary of errors (keyed by source) listing any problems
        encountered during parsing. Typical entries include validation errors
        and failures to provide values where required.  Raises a ValueError if
        ``parse()`` has not been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on this filter before reading errors")
        return self._errors

    @property
    def Q(self):
        """
        A Django "Q" object, built by combining input data with this filter's
        definition. Cannot be read before ``parse()`` has been called.
        """
        if not self.parsed:
            raise ValueError(
                "Must call parse() on this filter before using Q")

        dicts = [deepcopy(self.dict)]

        # Django, via SQL, does not do what you might expect with
        # .filter(rank__in=[1, 2, None]). This doesn't give you things with
        # rank 1, 2 or null -- you just get things with rank 1 and 2. Instead
        # SQL wants you to make an explicit "in or is null" check. For
        # consistency with Django and SQL, filternaut also behaves this way by
        # default. If you don't want this behaviour, argue none_to_isnull=True
        # when making a filter.
        if self.none_to_isnull:
            for key, val in list(dicts[0].items()):
                is_many = key.endswith('__in')
                has_null = is_many and None in val
                if is_many and has_null:
                    lookup = '{}__isnull'.format(key[:-4])
                    dicts.append({lookup: True})
                    val.remove(None)
                    # if the only value was None, we don't need the "__in" any
                    # more.
                    if not val:
                        dicts[0].pop(key)

        q = six.moves.reduce(or_, (Q(**d) for d in dicts))
        return ~q if self.negate else q



# -*- coding: utf8 -*-

from __future__ import unicode_literals, print_function, absolute_import

import json

from django.contrib.auth.models import User
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest
from django.test import TestCase, Client

from filternaut.filters import CharFilter, ChoiceFilter, EmailFilter


try:
    from django.test import RequestFactory
except ImportError:
    class RequestFactory(Client):
        """
        Django 1.2 does not have RequestFactory. credit:
        https://djangosnippets.org/snippets/963/
        """
        def request(self, **request):
            """
            Similar to parent class, but returns the request object as soon as
            it has created it.
            """
            environ = {
                'HTTP_COOKIE': self.cookies,
                'PATH_INFO': '/',
                'QUERY_STRING': '',
                'REQUEST_METHOD': 'GET',
                'SCRIPT_NAME': '',
                'SERVER_NAME': 'testserver',
                'SERVER_PORT': 80,
                'SERVER_PROTOCOL': 'HTTP/1.1',
            }
            environ.update(self.defaults)
            environ.update(request)
            return WSGIRequest(environ)


def user_to_native(user):
    return dict(
        id=user.id,
        username=user.username,
        email=user.email,
        first_name=user.first_name)


def resp_to_users(resp):
    data = json.loads(resp.content.decode('utf-8'))
    pks = [d['id'] for d in data]
    users = []
    for pk in pks:
        # get+append to preserve order of resp'd data
        users.append(User.objects.get(pk=pk))
    return users


def my_view(request):
    names = 'nomatch', 'Bret', 'Jemaine', 'Murray'
    filters = (
        CharFilter('username') | (
            EmailFilter('email') &
            ChoiceFilter('first_name', choices=zip(names, names))))
    filters.parse(request.GET)

    if filters.valid:
        users = User.objects.filter(filters.Q)
        native = [user_to_native(u) for u in users]
        return HttpResponse(json.dumps(native))
    else:
        return HttpResponseBadRequest(json.dumps(filters.errors))


class FullStackTests(TestCase):

    def setUp(self):
        self.factory = RequestFactory()
        self.u1 = User.objects.create(
            username='one',
            email='one@example.org',
            first_name='Bret')
        self.u2 = User.objects.create(
            username='two',
            email='two@example.org',
            first_name='Jemaine')
        self.u3 = User.objects.create(
            username='three',
            email='three@example.org',
            first_name='Murray')

    def test_request_no_params(self):
        """
        With no required filters, we present no filtering arguments and get all
        users returned.
        """
        request = self.factory.get('/my_view/')
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = [self.u1, self.u2, self.u3]

        assert resp.status_code == 200
        assert actual == expected

    def test_request_irrelevant_params(self):
        """
        Filternaut doesn't care if the 'foo' param is present
        """
        request = self.factory.get('/my_view/?foo=bar')
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = [self.u1, self.u2, self.u3]

        assert resp.status_code == 200
        assert actual == expected

    def test_request_nonmatching_param(self):
        """
        Filter on a non-existing username. Nothing comes back.
        """
        data = {'username': 'nomatch'}
        request = self.factory.get('/my_view/', data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = []

        assert resp.status_code == 200
        assert actual == expected

        assert resp.status_code == 200
        assert actual == expected

    def test_both_ANDed_filters_but_nonmatching(self):
        """
        Provide both ANDed filters, but with values that don't match anything.
        Expect no results.
        """
        data = {
            'email': 'nomatch@example.org',
            'first_name': 'nomatch'}
        request = self.factory.get('/my_view/', data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = []

        assert resp.status_code == 200
        assert actual == expected

    def test_both_ANDed_filters_mixed_matching(self):
        """
        Provide both ANDed filters, with a matching email, but non-matching
        first name. Expect no results.
        """
        data = {
            'email': 'one@example.org',  # a match
            'first_name': 'nomatch'}  # a non-match
        request = self.factory.get('/my_view/', data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = []

        assert resp.status_code == 200
        assert actual == expected

    def test_both_ANDed_filters_matching(self):
        """
        Provide both ANDed filters. Both are matching. Expect one matching
        result.
        """
        data = {
            'email': 'one@example.org',
            'first_name': 'Bret'}
        request = self.factory.get('/my_view/', data=data)
        resp = my_view(request)
        actual = resp_to_users(resp)
        expected = [self.u1]

        assert resp.status_code == 200
        assert actual == expected

    def test_invalid_filters(self):
        """
        Get some validation errors.
        """
        data = {
            'email': 'not an email',
            'first_name': 'not a choice'}
        request = self.factory.get('/my_view/', data=data)
        resp = my_view(request)
        errors = json.loads(resp.content.decode('utf-8'))

        assert resp.status_code == 400
        assert 'email' in errors
        assert 'first_name' in errors



# -*- coding: utf8 -*-

from __future__ import absolute_import, unicode_literals

from collections.abc import Iterable

import six

from django.forms import (BooleanField, CharField, ChoiceField, ComboField,
                          DateField, DateTimeField, DecimalField, EmailField,
                          FilePathField, FloatField, ImageField, IntegerField,
                          MultipleChoiceField, MultiValueField,
                          NullBooleanField, RegexField, SlugField,
                          SplitDateTimeField, TimeField, TypedChoiceField,
                          URLField)
from filternaut import Filter

# note IPAddressField, GenericIPAddressField and TypedMultipleChoiceField are
# conditionally imported later in this file; they are not available in all
# versions of Django.


__all__ = [
    'BooleanFilter', 'CharFilter', 'ChoiceFilter', 'ComboFilter', 'DateFilter',
    'DateTimeFilter', 'DecimalFilter', 'EmailFilter', 'FilePathFilter',
    'FloatFilter', 'IPAddressFilter', 'ImageFilter', 'FieldFilter',
    'IntegerFilter', 'MultiValueFilter', 'MultipleChoiceFilter',
    'NullBooleanFilter', 'RegexFilter', 'SlugFilter', 'SplitDateTimeFilter',
    'TimeFilter', 'TypedChoiceFilter', 'URLFilter']


def is_listlike(val):
    """
    True if `val` is an iterable (list, tuple, ...) but not a string
    """
    return isinstance(val, Iterable) and not isinstance(val, six.string_types)


class FieldFilter(Filter):
    """
    FieldFilters use a django.forms.Field to clean their input value when
    generating a Q object.

    This class is designed to be extended by subclasses which provide their own
    form-field instances. However, you could use it in combination with a
    custom field like so:

        filter = FieldFilter(SpecialField(), dest='...')
    """

    def __init__(self, dest, field, **kwargs):
        self.field = field
        super(FieldFilter, self).__init__(dest, **kwargs)

    def clean(self, value):
        if is_listlike(value):
            return type(value)(self.field.clean(v) for v in value)
        else:
            return self.field.clean(value)


# -- mixtures of fieldfilter and django fields requiring additional arguments


class ChoiceFilter(FieldFilter):
    def __init__(self, dest, choices, *args, **kwargs):
        field = ChoiceField(choices=choices)
        super(ChoiceFilter, self).__init__(dest, field=field, *args, **kwargs)


class RegexFilter(FieldFilter):
    def __init__(self, dest, regex, *args, **kwargs):
        field = RegexField(regex=regex)
        super(RegexFilter, self).__init__(dest, field=field, *args, **kwargs)


class FilePathFilter(FieldFilter):
    def __init__(self, dest, path, *args, **kwargs):
        field = FilePathField(path=path)
        super(FilePathFilter, self).__init__(dest, field=field,
                                             *args, **kwargs)


class BooleanFilter(FieldFilter):
    """
    BooleanField required=True does not have sanely.
    """
    def __init__(self, dest, **kwargs):
        field = BooleanField(required=False)
        super(BooleanFilter, self).__init__(dest, field=field, **kwargs)


class ComboFilter(FieldFilter):
    def __init__(self, dest, fields, **kwargs):
        field = ComboField(fields=fields)
        super(ComboFilter, self).__init__(dest, field=field, **kwargs)


# -- simple mixtures of fieldfilter and django fields.

class FieldMixin(object):

    def __init__(self, dest, **kwargs):
        field = self.field_class()
        super(FieldMixin, self).__init__(dest, field=field, **kwargs)


class CharFilter(FieldMixin, FieldFilter):
    field_class = CharField


class DateFilter(FieldMixin, FieldFilter):
    field_class = DateField


class DateTimeFilter(FieldMixin, FieldFilter):
    field_class = DateTimeField


class DecimalFilter(FieldMixin, FieldFilter):
    field_class = DecimalField


class EmailFilter(FieldMixin, FieldFilter):
    field_class = EmailField


class FloatFilter(FieldMixin, FieldFilter):
    field_class = FloatField


class ImageFilter(FieldMixin, FieldFilter):
    field_class = ImageField


class IntegerFilter(FieldMixin, FieldFilter):
    field_class = IntegerField


class MultiValueFilter(FieldMixin, FieldFilter):
    field_class = MultiValueField


class MultipleChoiceFilter(FieldMixin, FieldFilter):
    field_class = MultipleChoiceField


class NullBooleanFilter(FieldMixin, FieldFilter):
    field_class = NullBooleanField


class SlugFilter(FieldMixin, FieldFilter):
    field_class = SlugField


class SplitDateTimeFilter(FieldMixin, FieldFilter):
    field_class = SplitDateTimeField


class TimeFilter(FieldMixin, FieldFilter):
    field_class = TimeField


class TypedChoiceFilter(FieldMixin, FieldFilter):
    field_class = TypedChoiceField


class URLFilter(FieldMixin, FieldFilter):
    field_class = URLField


# -- mixtures of fieldfilter and django fields which were not present in older
# -- django versions

try:
    # django 1.3 and older did not have GenericIPAddressField or
    # TypedMultipleChoiceField, so we must conditionally define our matching
    # Filters.
    from django.forms import GenericIPAddressField, TypedMultipleChoiceField
except ImportError:
    pass
else:
    class GenericIPAddressFilter(FieldMixin, FieldFilter):
        field_class = GenericIPAddressField

    class TypedMultipleChoiceFilter(FieldMixin, FieldFilter):
        field_class = TypedMultipleChoiceField

    __all__.extend((
        'GenericIPAddressFilter',
        'TypedMultipleChoiceFilter'))

try:
    # django 1.9 and later drop support for IPAddressField
    from django.forms import IPAddressField
except ImportError:
    pass
else:
    class IPAddressFilter(FieldMixin, FieldFilter):
        field_class = IPAddressField

    __all__.extend(('IPAddressField', ))



# -*- coding: utf8 -*-

from __future__ import unicode_literals, print_function, absolute_import
from unittest import TestCase

from django.http import QueryDict

from filternaut.filters import Filter, FieldFilter


class ExtensionTests(TestCase):
    """
    Check that folk can extend filternaut in useful ways.
    """

    def test_with_custom_field(self):
        """
        Custom fields should work with FieldFilter.
        """
        class DoublingField(object):
            def clean(self, value):
                return int(value) * 2

        filter = FieldFilter('integer', field=DoublingField())
        filter.parse({'integer': 3})

        actual = dict(filter.Q.children)
        expected = {'integer': 6}
        assert actual == expected

    def test_basic_filter_with_subclassed_clean(self):
        """
        Basic filters should let their source values be cleaned.
        """
        class PotatoFilter(Filter):
            def clean(self, value):
                return 'potato'

        filter = PotatoFilter('fieldname')
        assert filter.clean('anything') == 'potato'

    def test_specialcase_multiple_source_values(self):
        """
        If multiple source values are presented to parse, make sure we allow
        handling the picking of a single value.
        """
        class BiggestValueFilter(Filter):
            def get_source_value(self, key, data, many=False):
                values = data.getlist(key)
                return max(map(int, values))
        data = QueryDict('field=3&field=6&field=1&field=2')
        filter = BiggestValueFilter('field')
        filter.parse(data)
        actual = dict(filter.Q.children)
        expected = {'field': 6}
        assert actual == expected



# -*- coding: utf8 -*-

from __future__ import unicode_literals, print_function, absolute_import
from operator import or_, and_

"""
This module ends up looking similar to django.utils.tree. It has a simpler
design allowed by the simpler use case. In particular there are
only ever two children, `left' and `right', which in turn allows the 'operator'
(what django terms a 'connector') to be literally a function from stdlib's
operator.

Finally, there are separate leaf and tree classes. This increases sanity when
these classes are extended to become filters.
"""


class Tree(object):

    # subclasses could specify a tree-subclass here. For example,
    # MyLeaf.tree_class = MyTree. A value of None will default to the current
    # class at the time of construction.
    tree_class = None

    def __init__(self, operator, left, right):
        assert operator in (or_, and_)
        self.operator = operator
        self.left = left
        self.right = right

    def __iter__(self):
        for l in self.left:
            yield l
        for r in self.right:
            yield r

    def __and__(self, other):
        klass = getattr(self, 'tree_class') or self.__class__
        return klass(operator=and_, left=self, right=other)

    def __or__(self, other):
        klass = getattr(self, 'tree_class') or self.__class__
        return klass(operator=or_, left=self, right=other)


class Leaf(object):

    # subclasses could specify a tree-subclass here. For example,
    # MyLeaf.tree_class = MyTree.
    tree_class = Tree

    def __iter__(self):
        yield self

    def __and__(self, other):
        return self.tree_class(operator=and_, left=self, right=other)

    def __or__(self, other):
        return self.tree_class(operator=or_, left=self, right=other)
